import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_left_child_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `left_child`, containing for each node its smallest-id child.

    Leaf nodes will be marked with their own id.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["left_child"] = phylogeny_df["id"]

    for idx in phylogeny_df.index:
        ancestor_id = phylogeny_df.loc[idx, "ancestor_id"]
        if ancestor_id == idx:
            continue  # handle genesis cases

        cur_left_child = phylogeny_df.loc[ancestor_id, "left_child"]
        if cur_left_child == ancestor_id or idx < cur_left_child:
            phylogeny_df.loc[ancestor_id, "left_child"] = idx

    return phylogeny_df



import pandas as pd
import tqdist

from . import (
    alifestd_as_newick_asexual,
    alifestd_collapse_unifurcations,
    alifestd_count_root_nodes,
    alifestd_mark_leaves,
)


# adapted from https://github.com/mmore500/hstrat/blob/d23917cf03ba59061ff2f9b951efe79e995eb4d8/tests/test_hstrat/test_phylogenetic_inference/test_tree/_impl/_tree_quartet_distance.py
def alifestd_calc_triplet_distance_asexual(
    ref: pd.DataFrame,
    cmp: pd.DataFrame,
    taxon_label_key: str = "taxon_label",
) -> float:
    """Calculate the triplet distance between two trees."""

    ref = alifestd_mark_leaves(alifestd_collapse_unifurcations(ref))
    cmp = alifestd_mark_leaves(alifestd_collapse_unifurcations(cmp))

    ref.loc[~ref["is_leaf"], taxon_label_key] = ""
    cmp.loc[~cmp["is_leaf"], taxon_label_key] = ""

    ref_labels = {*ref[taxon_label_key][ref["is_leaf"]]}
    cmp_labels = {*cmp[taxon_label_key][cmp["is_leaf"]]}

    if ref_labels != cmp_labels:
        raise ValueError("Taxon labels must match between trees")
    for taxon_label in ref_labels:
        if isinstance(taxon_label, str) and not taxon_label.strip():
            raise ValueError("Cannot have empty taxon labels")

    if (
        alifestd_count_root_nodes(ref) > 1
        or alifestd_count_root_nodes(cmp) > 1
    ):
        raise ValueError(
            f"Cannot have disjunct trees in `{alifestd_calc_triplet_distance_asexual.__name__}`"
        )

    ref_newick = (
        alifestd_as_newick_asexual(ref, taxon_label=taxon_label_key)
        .removeprefix("[&R]")
        .strip()
    )
    cmp_newick = (
        alifestd_as_newick_asexual(cmp, taxon_label=taxon_label_key)
        .removeprefix("[&R]")
        .strip()
    )
    return tqdist.triplet_distance(ref_newick, cmp_newick)



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted


def alifestd_is_working_format_asexual(
    phylogeny_df,
    mutate: bool = False,
) -> pd.DataFrame:
    """Test if phylogeny_df is an asexual phylogeny in working format.

    The working format is a dataframe with the following properties:
      - topologically sorted (i.e., organisms appear after all ancestors),
      - contiguous ids (i.e., organisms' ids correspond to row number), and
      - contains an integer datatype `ancestor_id` column.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    return (
        "ancestor_id" in phylogeny_df.columns
        and alifestd_is_topologically_sorted(phylogeny_df)
        and alifestd_has_contiguous_ids(phylogeny_df)
    )



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_left_child_asexual import alifestd_mark_left_child_asexual
from ._alifestd_mark_roots import alifestd_mark_roots
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_is_left_child_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `is_left_child`, containing for each node whether it is the
    smaller-id child.

    Root nodes will be marked False. Tree must be strictly bifurcating.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    if "left_child" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_left_child_asexual(
            phylogeny_df, mutate=True
        )

    if "is_root" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)

    phylogeny_df["is_left_child"] = (
        phylogeny_df.loc[phylogeny_df["ancestor_id"], "left_child"].values
        == phylogeny_df["id"].values
    ) & ~phylogeny_df["is_root"].values

    return phylogeny_df



import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_mark_num_children_asexual import (
    alifestd_mark_num_children_asexual,
)
from ._alifestd_mark_roots import alifestd_mark_roots
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_delete_unifurcating_roots_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    root_ancestor_token: str = "none",
) -> pd.DataFrame:
    """Pare record to bypass root nodes with only one descendant.

    May leave a root unifurcation present. See
    `alifestd_collapse_root_unifurcations_asexual`.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_num_children_asexual(
        phylogeny_df, mutate=True
    )
    phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    assert "ancestor_id" in phylogeny_df

    phylogeny_df["is_unifurcating_root"] = (
        phylogeny_df["num_children"] == 1
    ) & phylogeny_df["is_root"]

    phylogeny_df.set_index("id", drop=False, inplace=True)
    phylogeny_df["ancestor_is_unifurcating_root"] = phylogeny_df.loc[
        phylogeny_df["ancestor_id"], "is_unifurcating_root"
    ].values

    phylogeny_df.loc[
        phylogeny_df["ancestor_is_unifurcating_root"],
        "ancestor_id",
    ] = phylogeny_df.loc[
        phylogeny_df["ancestor_is_unifurcating_root"],
        "id",
    ].values

    phylogeny_df = phylogeny_df[~phylogeny_df["is_unifurcating_root"]].copy()

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[:, "ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"],
            phylogeny_df["ancestor_id"],
            root_ancestor_token=root_ancestor_token,
        )

    return phylogeny_df



from ._auxiliary_lib import get_hstrat_version

if __name__ == "__main__":
    print(f"hstrat v{get_hstrat_version()}")
    print(
        """
Available commands (stabilized API):
$ python3 -m hstrat.dataframe.surface_build_tree
$ python3 -m hstrat.dataframe.surface_unpack_reconstruct
$ python3 -m hstrat.dataframe.surface_postprocess_trie

Available commands (experimental API):
$ python3 -m hstrat._auxiliary_lib._alifestd_as_newick_asexual
$ python3 -m hstrat._auxiliary_lib._alifestd_count_root_nodes
$ python3 -m hstrat._auxiliary_lib._alifestd_downsample_tips_asexual
$ python3 -m hstrat._auxiliary_lib._alifestd_join_roots
$ python3 -m hstrat._auxiliary_lib._alifestd_test_leaves_isomorphic_asexual
$ python3 -m hstrat._auxiliary_lib._alifestd_to_working_format
$ python3 -m hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col

For information on a command, invoke it with the --help flag.
""",
    )



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_mark_is_right_child_asexual import (
    alifestd_mark_is_right_child_asexual,
)
from ._alifestd_mark_num_leaves_asexual import alifestd_mark_num_leaves_asexual
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_num_preceding_leaves_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `num_preceding_leaves` with count of all leaves occuring
    before the present node in an inorder traversal.

    For internal nodes, the number of leaf nodes prior to the traversal of
    first (i.e., leftmost) descendant is marked.

    A topological sort will be applied if `phylogeny_df` is not topologically
    sorted. Dataframe reindexing (e.g., df.index) may be applied.

    Must be a strictly bifurcating tree.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    if "num_leaves" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_leaves_asexual(
            phylogeny_df, mutate=True
        )

    if "is_right_child" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_is_right_child_asexual(
            phylogeny_df, mutate=True
        )

    phylogeny_df["num_preceding_leaves"] = 0

    for idx in phylogeny_df.index:
        ancestor_id = phylogeny_df.loc[idx, "ancestor_id"]
        phylogeny_df.loc[idx, "num_preceding_leaves"] = (
            phylogeny_df.loc[ancestor_id, "num_preceding_leaves"]
            + (
                phylogeny_df.loc[ancestor_id, "num_leaves"]
                - phylogeny_df.loc[idx, "num_leaves"]
            )
            * phylogeny_df.loc[idx, "is_right_child"]
        )

    return phylogeny_df



import logging
import numbers
import typing
import warnings

import numpy as np
import opytional as opyt
import pandas as pd
import polars as pl


def alifestd_prefix_roots_polars(
    phylogeny_df: pd.DataFrame,
    *,
    allow_id_reassign: bool = False,
    origin_time: typing.Optional[numbers.Real] = None,
) -> pd.DataFrame:
    """Add new roots to the phylogeny, prefixing existing roots.

    An origin time may be specified, in which case only roots with origin times
    past the specified time will be prefixed. If no origin time is specified,
    all roots will be prefixed.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if "origin_time_delta" in phylogeny_df:
        warnings.warn("alifestd_prefix_roots ignores origin_time_delta values")
    if origin_time is not None and "origin_time" not in phylogeny_df:
        raise ValueError(
            "origin_time specified but not present in phylogeny dataframe",
        )

    if "ancestor_list" in phylogeny_df:
        raise NotImplementedError
    if not allow_id_reassign:
        raise NotImplementedError
    if phylogeny_df.is_empty():
        raise NotImplementedError
    has_contiguous_ids = phylogeny_df.select(
        pl.col("id").diff() == 1
    ).to_series().all() and (phylogeny_df["id"].first() == 0)
    if not has_contiguous_ids:
        raise NotImplementedError

    phylogeny_df = phylogeny_df.drop("is_root", strict=False)

    logging.info("- alifestd_prefix_roots: marking eligible roots...")
    eligible_roots = (
        phylogeny_df.lazy()
        .with_columns(
            is_eligible=pl.col("origin_time") > origin_time
            if origin_time is not None
            else True,
            is_root=pl.col("id") == pl.col("ancestor_id"),
        )
        .select(
            pl.col("is_eligible") & pl.col("is_root"),
        )
        .collect()
        .to_series()
    )

    logging.info("- alifestd_prefix_roots: filtering prepended roots...")
    prepended_roots = (
        phylogeny_df.lazy()
        .filter(
            eligible_roots,
        )
        .select(
            "id",
            "origin_time",
            "ancestor_id",
        )
        .collect()
    )

    if "origin_time" in prepended_roots:
        logging.info("- alifestd_prefix_roots: setting origin time...")
        prepended_roots = prepended_roots.with_columns(
            origin_time=pl.lit(opyt.or_value(origin_time, 0))
        )

    logging.info("- alifestd_prefix_roots: updating phylogeny_df...")
    phylogeny_df = phylogeny_df.with_columns(
        id=pl.col("id") + pl.lit(len(prepended_roots)),
        ancestor_id=pl.col("ancestor_id") + pl.lit(len(prepended_roots)),
    )
    ancestor_ids = phylogeny_df["ancestor_id"].to_numpy().copy()
    ancestor_ids[prepended_roots["id"].to_numpy()] = np.arange(
        len(prepended_roots),
    )
    phylogeny_df = phylogeny_df.with_columns(
        ancestor_id=pl.Series(ancestor_ids),
    )

    logging.info("- alifestd_prefix_roots: updating prepended roots...")
    prepended_roots = prepended_roots.with_columns(
        id=pl.int_range(len(prepended_roots)),
        ancestor_id=pl.int_range(len(prepended_roots)),
    ).cast(
        {
            k: v
            for k, v in phylogeny_df.collect_schema().items()
            if k in prepended_roots.collect_schema()
        },
    )

    logging.info("- alifestd_prefix_roots: creating gather_indices...")
    gather_indices = np.empty(
        len(phylogeny_df) + len(prepended_roots), dtype=np.int64
    )
    gather_indices[: len(prepended_roots)] = np.arange(len(prepended_roots))
    gather_indices[len(prepended_roots) :] = np.arange(len(phylogeny_df))

    logging.info("- alifestd_prefix_roots: gathering and updating...")
    return (
        phylogeny_df.lazy()
        .select(pl.all().gather(gather_indices))
        .with_row_index()
        .with_columns(
            pl.when(pl.col("index") < len(prepended_roots))
            .then(None)
            .otherwise(pl.all())
            .name.keep()
        )
        .drop("index")
        .update(prepended_roots.lazy())
        .collect()
    )



import os
import pathlib
import subprocess

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_alifestd_downsample_tips_asexual_cli_help():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_downsample_tips_asexual",
            "--help",
        ],
        check=True,
    )


def test_alifestd_downsample_tips_asexual_cli_version():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_downsample_tips_asexual",
            "--version",
        ],
        check=True,
    )


def test_alifestd_downsample_tips_asexual_cli_csv():
    output_file = "/tmp/hstrat_alifestd_downsample_tips_asexual.csv"
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_downsample_tips_asexual",
            "-n",
            "1",
            output_file,
        ],
        check=True,
        input=f"{assets}/example-standard-toy-asexual-phylogeny.csv".encode(),
    )
    assert os.path.exists(output_file)


def test_alifestd_downsample_tips_asexual_cli_parquet():
    output_file = "/tmp/hstrat_alifestd_downsample_tips_asexual.pqt"
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_downsample_tips_asexual",
            "-n",
            "1",
            "--seed",
            "50_000_000",
            output_file,
        ],
        check=True,
        input=f"{assets}/example-standard-toy-asexual-phylogeny.csv".encode(),
    )
    assert os.path.exists(output_file)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_calc_triplet_distance_asexual,
    alifestd_make_ancestor_list_col,
)


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "taxon_label": [0, 1, 2, 3],
                "ancestor_id": [0, 0, 0, 0],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5],
                "taxon_label": ["0", "1", "2", "3", "4", "5"],
                "ancestor_id": [0, 0, 0, 0, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [9, 1, 2, 3, 4, 5],
                "taxon_label": ["9", "1", "2", "3", "4", "5"],
                "ancestor_id": [9, 9, 9, 9, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5, 6],
                "taxon_label": ["0", "1", "2", "3", "4", "5", "6"],
                "ancestor_id": [0, 0, 0, 0, 1, 2, 3],
            },
        ),
    ],
)
def test_polytomy_identical(df: pd.DataFrame):
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for adf, bdf in (df, df), (df, df.sample(frac=1)):
        assert alifestd_calc_triplet_distance_asexual(adf, bdf) == 0


def test_differing_wrong1():
    adf = pd.DataFrame(
        {
            "id": reversed([9, 1, 2, 3, 4, 5]),
            "taxon_label": reversed(["0", "1", "2", "3", "4", "5"]),
            "ancestor_id": reversed([9, 9, 1, 2, 2, 1]),
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 5, 4],
            "taxon_label": ["0", "1", "2", "3", "5", "4"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    assert 0 < alifestd_calc_triplet_distance_asexual(adf, bdf) < 1
    assert (
        0 < alifestd_calc_triplet_distance_asexual(adf, bdf.sample(frac=1)) < 1
    )


def test_differing_wrong2():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 1, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    assert 0 < alifestd_calc_triplet_distance_asexual(adf, bdf) < 1
    assert (
        0 < alifestd_calc_triplet_distance_asexual(adf, bdf.sample(frac=1)) < 1
    )
    assert (
        0
        < alifestd_calc_triplet_distance_asexual(
            adf.sample(frac=1), bdf.sample(frac=1)
        )
        < 1
    )


def test_identical_polytomy1():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 3, 2],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 3, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    assert alifestd_calc_triplet_distance_asexual(adf, bdf, "taxon_label") == 0
    assert (
        alifestd_calc_triplet_distance_asexual(
            adf, bdf.sample(frac=1), "taxon_label"
        )
        == 0
    )


def test_differing_wrong_big():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 5, 5, 4, 4],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 5, 2, 1, 2, 5, 4, 4],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_calc_triplet_distance_asexual(
        adf,
        bdf,
        "taxon_label",
    )
    assert 0 < est < 1

    est = alifestd_calc_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
    )
    assert 0 < est < 1


def test_differing_polytomy_asymmetrical_strict():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    assert (
        0 < alifestd_calc_triplet_distance_asexual(adf, bdf, "taxon_label") < 1
    )
    assert (
        0 < alifestd_calc_triplet_distance_asexual(adf, bdf, "taxon_label") < 1
    )
    assert (
        0
        < alifestd_calc_triplet_distance_asexual(
            adf,
            bdf.sample(frac=1),
            "taxon_label",
        )
        < 1
    )
    assert (
        0
        < alifestd_calc_triplet_distance_asexual(
            adf,
            bdf.sample(frac=1),
            "taxon_label",
        )
        < 1
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_collapse_unifurcations,
    alifestd_make_empty,
    alifestd_test_leaves_isomorphic_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    mt = alifestd_make_empty()
    mt["taxon_label"] = None
    assert alifestd_test_leaves_isomorphic_asexual(mt, mt, "taxon_label")


def test_mutate():
    original_df = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    original_df["taxon_label"] = original_df["id"]
    df = original_df.copy()
    alifestd_test_leaves_isomorphic_asexual(
        df, df, "taxon_label", mutate=False
    )
    assert df.equals(original_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz_positive1(phylogeny_df: pd.DataFrame):

    phylogeny_df["taxon_label"] = phylogeny_df["id"]
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df, phylogeny_df, "taxon_label"
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df,
        alifestd_collapse_unifurcations(phylogeny_df),
        "taxon_label",
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df,
        phylogeny_df.sample(frac=1).reset_index(drop=True),
        "taxon_label",
    )


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz_positive2(phylogeny_df: pd.DataFrame):

    phylogeny_df["taxon_label"] = phylogeny_df["id"].astype(str)
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df, phylogeny_df, "taxon_label"
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df,
        alifestd_collapse_unifurcations(phylogeny_df),
        "taxon_label",
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylogeny_df,
        phylogeny_df.sample(frac=1).reset_index(drop=True),
        "taxon_label",
    )


def test_negative1():
    df1 = pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
    df2 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    df1["taxon_label"] = df1["id"]
    df2["taxon_label"] = df2["id"]
    assert not alifestd_test_leaves_isomorphic_asexual(df1, df2, "taxon_label")


def test_negative2():
    df1 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    df2 = pd.read_csv(f"{assets_path}/nk_ecoeaselection_tweaked.csv")
    df1["taxon_label"] = df1["id"].astype(str)
    df2["taxon_label"] = df2["id"].astype(str)
    assert not alifestd_test_leaves_isomorphic_asexual(df1, df2, "taxon_label")



import logging
import warnings

import pandas as pd
import polars as pl

from ._alifestd_assign_contiguous_ids_polars import (
    alifestd_assign_contiguous_ids_polars,
)
from ._alifestd_collapse_unifurcations import _collapse_unifurcations


def alifestd_collapse_unifurcations_polars(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Pare record to bypass organisms with one ancestor and one descendant.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if any(
        col in phylogeny_df
        for col in ["branch_length", "edge_length", "origin_time_delta"]
    ):
        warnings.warn(
            "alifestd_collapse_unifurcations does not update branch length "
            "columns. Use `origin_time` to recalculate branch lengths for "
            "collapsed phylogeny."
        )

    if "ancestor_list" in phylogeny_df:
        raise NotImplementedError

    original_ids = phylogeny_df["id"]
    has_contiguous_ids = phylogeny_df.select(
        pl.col("id").diff() == 1
    ).to_series().all() and (phylogeny_df["id"].first() == 0)
    if not has_contiguous_ids:
        logging.info(
            "- alifestd_collapse_unifurcations_polars: assigning ids...",
        )
        phylogeny_df = alifestd_assign_contiguous_ids_polars(phylogeny_df)

    if (
        not phylogeny_df.select(pl.col("ancestor_id") <= pl.col("id"))
        .to_series()
        .all()
    ):
        raise NotImplementedError(
            "polars topological sort not yet implemented"
        )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: calculating reindex...",
    )
    keep_filter, ancestor_ids = _collapse_unifurcations(
        # must copy to remove read-only flag for numba compatibility
        phylogeny_df["ancestor_id"]
        .to_numpy()
        .copy(),
    )

    logging.info(
        "- alifestd_collapse_unifurcations_polars: applying reindex...",
    )
    return phylogeny_df.filter(keep_filter).with_columns(
        id=original_ids.to_numpy()[keep_filter],
        ancestor_id=original_ids.gather(ancestor_ids[keep_filter]),
    )



import operator

import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_lineage_asexual import alifestd_unfurl_lineage_asexual
from ._jit import jit
from ._unfurl_lineage_with_contiguous_ids import (
    unfurl_lineage_with_contiguous_ids,
)


def _create_has_extant_descendant_noncontiguous(
    phylogeny_df: pd.DataFrame,
    extant_mask: np.ndarray,
) -> np.ndarray:
    """Implementation detail for alifestd_prune_extinct_lineages_asexual."""

    phylogeny_df["has_extant_descendant"] = False
    for extant_id in phylogeny_df.loc[extant_mask, "id"]:
        for lineage_id in alifestd_unfurl_lineage_asexual(
            phylogeny_df,
            int(extant_id),
            mutate=True,
        ):
            if phylogeny_df.loc[lineage_id, "has_extant_descendant"]:
                break

            phylogeny_df.loc[lineage_id, "has_extant_descendant"] = True

    return phylogeny_df["has_extant_descendant"]


@jit(nopython=True)
def _create_has_extant_descendant_contiguous(
    ancestor_ids: np.ndarray,
    extant_mask: np.ndarray,
) -> np.ndarray:
    """Implementation detail for alifestd_prune_extinct_lineages_asexual."""

    has_extant_descendant = np.zeros_like(extant_mask)
    for extant_id in np.flatnonzero(extant_mask):
        for lineage_id in unfurl_lineage_with_contiguous_ids(
            ancestor_ids,
            int(extant_id),
        ):
            if has_extant_descendant[lineage_id]:
                break

            has_extant_descendant[lineage_id] = True

    return has_extant_descendant


@jit(nopython=True)
def _create_has_extant_descendant_contiguous_sorted(
    ancestor_ids: np.ndarray,
    extant_mask: np.ndarray,
) -> np.ndarray:
    """Implementation detail for alifestd_prune_extinct_lineages_asexual."""

    has_extant_descendant = extant_mask.copy()
    for id_ in range(len(ancestor_ids) - 1, -1, -1):
        has_extant_descendant[ancestor_ids[id_]] |= has_extant_descendant[id_]

    return has_extant_descendant


def alifestd_prune_extinct_lineages_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Drop taxa without extant descendants.

    An "extant" column, if provided, is used to determine extant taxa.
    Otherwise, taxa with inf or nan "destruction_time" are considered extant.

    Fastest with records in working format. See `alifestd_to_working_format`.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    mutate : bool, default False
        Are side effects on the input argument `phylogeny_df` allowed?

    Raises
    ------
    ValueError
        If `phylogeny_df` has neither "extant" or "destruction_time" columns.

        Without at least one of these columns, which taxa are extant is
        ambiguous.

    Returns
    -------
    pandas.DataFrame
        The rerooted phylogeny in alife standard format.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    extant_mask = None
    if "extant" in phylogeny_df:
        extant_mask = phylogeny_df["extant"]
    elif "destruction_time" in phylogeny_df:
        extant_mask = operator.or_(
            phylogeny_df["destruction_time"].isna(),
            np.isinf(phylogeny_df["destruction_time"]),
        )
    else:
        raise ValueError('Need "extant" or "destruction_time" column.')

    if not alifestd_has_contiguous_ids(phylogeny_df):
        has_extant_descendant = _create_has_extant_descendant_noncontiguous(
            phylogeny_df,
            extant_mask,
        )
    elif not alifestd_is_topologically_sorted(phylogeny_df):
        has_extant_descendant = _create_has_extant_descendant_contiguous(
            phylogeny_df["ancestor_id"].to_numpy(dtype=np.uint64),
            extant_mask.to_numpy(dtype=bool),
        )
    else:
        has_extant_descendant = (
            _create_has_extant_descendant_contiguous_sorted(
                phylogeny_df["ancestor_id"].to_numpy(dtype=np.uint64),
                extant_mask.to_numpy(dtype=bool),
            )
        )

    phylogeny_df = phylogeny_df[has_extant_descendant].reset_index(drop=True)
    phylogeny_df.drop(
        columns="has_extant_descendant", errors="ignore", inplace=True
    )
    return phylogeny_df
